/////////////////////////////////////////////////////////////////////
//
// RC Library, (c) 2011-2014, Ryan A. Colyer
// Distributed under the Boost Software License, v1.0. (LICENSE.txt)
//
/// \file Net.h
/// Provides basic socket networking.
/////////////////////////////////////////////////////////////////////

#ifndef RC_NET_H
#define RC_NET_H

#include "Types.h"
#include "Errors.h"
#include "File.h"
#include "RStr.h"
#include <string.h>
#ifdef WIN32
#ifndef _WIN32_WINNT
/// @cond UNDOC
#define _WIN32_WINNT 0x0501  // XP and later only
/// @endcond
#endif
#include <winsock2.h>
#include <ws2tcpip.h>
#else
#include <sys/socket.h>
#include <netdb.h>
#endif
#include <sys/types.h>


namespace RC {
  /// Provides both client and server sides of blocking TCP connections.
  class Net {
    public:

    /// Automatically called by Connect and Listener.
    /** The first time this is called, it sets SIGPIPE to ignore so the
     *  program doesn't terminate while writing to a closed socket.  Call
     *  this once manually beforehand if you want to set your own custom
     *  handler for SIGPIPE.
     */
    static inline void InitializeSockets() {
      static RC_THREAD_LOCAL bool already_run = false;
      if (! already_run) {
#ifndef WIN32
        struct sigaction sa;
        memset (&sa, 0, sizeof(sa));
        sa.sa_handler = SIG_IGN;
        sigaction(SIGPIPE, &sa, NULL);

#endif

#ifdef WIN32
        WSADATA wsa_data;
        int err;
        err = WSAStartup(0x0202, &wsa_data);
        if (err != 0) {
          Throw_RC_Type(Net, "WSAStartup error");
        }
#endif

        already_run = true;
      }
    }


    /// Opens a TCP connection to host:port.
    /** Will throw an exception upon failure to connect if quiet_fail is
     *  false.  The FileRW connection provides blocking access to read or
     *  write to the socket.
     *  @return True if the connect succeeded.
     */
    static inline bool Connect(FileRW &connection, const RStr& host,
                               const RStr& port, bool quiet_fail=true) {
      FileRW retval;
      struct addrinfo hints;
      struct addrinfo *ai_list, *addr;
      int sfd;

      InitializeSockets();

      memset(&hints, 0, sizeof(hints));
      hints.ai_family = AF_UNSPEC;  // Support IPv4 && IPv6
      hints.ai_socktype = SOCK_STREAM;  // For TCP

      int aierr = getaddrinfo(host.c_str(), port.c_str(), &hints, &ai_list);
      if (aierr != 0) {
        if (! quiet_fail) {
          Throw_RC_Type(Net, RStr(gai_strerror(aierr)).c_str());
        }
        return false;
      }
      
      for (addr = ai_list; addr != NULL; addr = addr->ai_next) {
        sfd = socket(addr->ai_family, addr->ai_socktype, addr->ai_protocol);
        if (sfd == -1) {
          continue;
        }

        if (connect(sfd, addr->ai_addr, addr->ai_addrlen) != -1) {
          break;  // Connected.
        }

        close(sfd);
      }

      // If could not find one
      if (addr == NULL) {
        if (! quiet_fail) {
          Throw_RC_Error("Connection failed");
        }
        return false;
      }

      freeaddrinfo(ai_list);

      FILE* sfp = fdopen(sfd, "a+");
      if (sfp == NULL) {
        close(sfd);
        if (! quiet_fail) {
          Throw_RC_Error("fdopen failed");
        }
      }


      connection = FileRW(sfp);
      connection.SetFilename(host + " " + port);
      return true;
    }


    /// Listens to the specified port for incoming TCP connections.
    class Listener {
      protected:
      /// @cond PROTECTED
      int sockfd;

      /// @endcond
      public:

      /// Begin listening to the specified port for incoming TCP connections.
      /** Use Accept to accept a connection.
       */
      inline Listener(const RStr& port) {
        sockfd = -1;

        struct addrinfo hints;
        struct addrinfo *ai_list, *addr;
        int sfd;

        InitializeSockets();

        memset(&hints, 0, sizeof(hints));
        hints.ai_family = AF_UNSPEC;  // Support IPv4 && IPv6
        hints.ai_socktype = SOCK_STREAM;  // For TCP
        hints.ai_flags = AI_PASSIVE;  // Listen on all interfaces

        int aierr = getaddrinfo(NULL, port.c_str(), &hints, &ai_list);
        if (aierr != 0) {
          Throw_RC_Error(RStr(gai_strerror(aierr)).c_str());
        }
        
        for (addr = ai_list; addr != NULL; addr = addr->ai_next) {
          sfd = socket(addr->ai_family, addr->ai_socktype, addr->ai_protocol);
          if (sfd == -1) {
            continue;
          }

          if (::bind(sfd, addr->ai_addr, addr->ai_addrlen) == 0) {
            break;  // Bound.
          }

          close(sfd);
        }

        // If could not find one
        if (addr == NULL) {
          Throw_RC_Error("Could not bind");
        }

        freeaddrinfo(ai_list);

        if (listen(sfd, 128) == -1) {
          Throw_RC_Error("Port in use");
        }

        sockfd = sfd;
      }


      /// If open, this closes the port being listened to.
      ~Listener() {
        if (sockfd != -1) {
          close(sockfd);
        }
      }


      /// Provides the raw socket descriptor being listened to.
      int Raw() const {
        return sockfd;
      }

      /// Accept an incoming connection, with read/write access through the
      /// FileRW connection.
      /** The filename of the FileRW is set to "hostname port".  This throws
       *  an ErrorMsgNet on error if quiet_fail is false.
       *  @return False if an error occurred.
       */
      inline bool Accept(FileRW &connection, bool quiet_fail=true) {
        int acc_fd;
        struct sockaddr_in6 addr;
        socklen_t len = sizeof(addr);

        acc_fd = accept(sockfd, (sockaddr*)&addr, &len);

        if (acc_fd == -1) {
          if (! quiet_fail) {
            Throw_RC_Type(Net, "Could not accept connection");
          }
          return false;
        }

        FILE* afp = fdopen(acc_fd, "a+");
        if (afp == NULL) {
          close(acc_fd);
          if (! quiet_fail) {
            Throw_RC_Type(Net, "fdopen failed");
          }
          return false;
        }

        connection.Open(afp);

        char host[2048];
        char port[2048];
        if (getnameinfo((sockaddr*)&addr, sizeof(addr), host, 2048, port, 2048,
                        NI_NUMERICHOST | NI_NUMERICSERV) == 0) {
          // Success
          connection.SetFilename(RStr(host) + " " + RStr(port));
        }

        return true;
      }
    };

  };
}


#endif // RC_NET_H

